(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import utils pos.

Require Import uterm uterm_beta uterm_beta_eq.
Require Import uterm_normal uterm_head uterm_solvable.
Require Import uterm_nat uterm_value.

Set Implicit Arguments.

Definition ut_comp n (t u : uterm n) := \\ (ut_lift t @ (ut_lift u @ £ pos0)).

Fact ut_comp_subst n m t u (s : uts_t n m) : ut_subst (@ut_comp n t u) s = ut_comp (ut_subst t s) (ut_subst u s).
Proof. 
  unfold ut_comp, ut_subst; fold ut_subst.
  repeat rewrite <- ut_lift_subst_2; auto.
Qed.

(* There is NO NEED for s and h to be closed HERE, contrary to what is expressed 
   in Lemma 2.5 p33 of Krivine's book ... did he formulate it this way to avoid 
   capture of bound variables ? Anyway the proof of Lemma 2.6 uses Lemma 2.5 for
   an unclosed term and is thus wrong ... Lemma 2.5 should be corrected in that
   book 
   
   The mistake is the following ... on line 3 of the proof of Lemma 2.5 p 33,
   it is written that by lemma 2.5 
      \h(\g g o suc)^n h head reduces to \h\x(h)(suc)^n x             (1)  
   But lemma 2.5 only states that
       (\g g o suc)^n h head reduces to \x(h)(suc)^n x                (2)
   for any CLOSED term h. So you cannot use it in (1) because under the lambda
   \h, h is a variable and hence is not closed !!
   
   Anyway, Lemma 2.5 can be proved for arbitrary h without any modification to
   the proof so I do not know why it is stated that way ?
*)

Fact ut_church_comp_prop n k (s h : uterm n) :
                                       ut_church (S k) (\\ ut_comp (£ (pos_fst _)) (ut_lift s)) h 
                                  -h>> \\ (ut_lift h @ ut_church (S k) (ut_lift s) (£ (pos_fst _))).
Proof.
  revert h; induction k as [ | k IHk ]; intros h.
  do 2 rewrite ut_church_S_tail.
  do 2 rewrite ut_church_O.
  ut_head_rt_step; ut_simpl; constructor 2.
  
  revert IHk.
  set (u := \\ ut_comp (£ (pos_fst _)) (ut_lift s)).
  intros IHk.
  rewrite ut_church_S_tail with (t := u).
  apply ut_head_rt_trans with (1 := IHk (u @ h)).
  unfold u; clear u IHk.
  repeat rewrite ut_lift_subst_eq.
  unfold ut_subst; fold ut_subst.
  rewrite ut_comp_subst.
  ut_simpl.
  ut_head_rt_step.
  ut_simpl.
  assert (forall (h : uterm n), ut_subst h (fun p => £ (pos_nxt _ (pos_nxt _ p))) = ut_lift (ut_lift h)) as H.
    intros;
    repeat rewrite ut_lift_subst_eq.
    rewrite ut_subst_comp; auto.
  repeat rewrite H.
  ut_head_rt_step.
  ut_simpl.
  clear H.
  assert (forall (h : uterm n), ut_subst h (fun p => £ ((pos_nxt _ p))) = (ut_lift h)) as H.
    intros;
    repeat rewrite ut_lift_subst_eq; auto.
  repeat rewrite H.
  rewrite <- ut_church_S_head.
  constructor 2.
Qed.
  
Section ut_apply.

  Variable (n : nat) (phi v : uterm n).
  
  Definition ut_apply := v @ \\ (ut_comp (£ pos0) (ut_succ _)) @ phi @ ut_nat n 0.

  Fact ut_apply_v_solvable : ut_solvable ut_apply -> ut_solvable v.
  Proof.
    intros H; do 3 apply ut_app_solvable in H; auto.
  Qed.

  Variable k : nat.
  Hypothesis Hk : v ~b ut_nat _ k.

  Let Hk' : v -b>> ut_nat _ k.
  Proof.
    apply ut_beta_normal_eq; auto.
    apply ut_nat_normal.
  Qed.

  Fact ut_apply_beta_eq : ut_apply ~b phi @ ut_nat _ k.
  Proof.
    unfold ut_apply.
    set (u := \\ ut_comp £(pos_fst _) (ut_succ _)).
    apply ut_beta_rt_lft with (a := u) in Hk'.
    apply ut_beta_rt_lft with (a := phi) in Hk'.
    apply ut_beta_rt_lft with (a := ut_nat _ 0) in Hk'.
    apply ut_beta_rt_eq in Hk'.
    apply ut_beta_eq_trans with (1 := Hk'); clear Hk Hk'.
    generalize (ut_nat_reduce k u phi); intros H.
    apply ut_beta_rt_lft with (a := ut_nat _ 0) in H.
    apply ut_beta_rt_eq in H.
    apply ut_beta_eq_trans with (1 := H); clear H.
    
    destruct k as [ | k' ].
    rewrite ut_church_O; constructor 2.
    generalize (@ut_church_comp_prop n k' (ut_succ _) phi); intros H.
    repeat rewrite ut_lift_subst_eq in H.
    repeat rewrite ut_succ_subst in H.
    repeat rewrite ut_nat_subst in H.
    fold u in H.
    apply ut_head_rt_beta_rt in H.
    apply ut_beta_rt_lft with (a := ut_nat _ 0) in H.
    apply ut_beta_rt_eq in H.
    apply ut_beta_eq_trans with (1 := H).
    clear H u k.
    apply ut_beta_rt_eq.
    ut_beta_rt_step.
    unfold ut_reduce.
    unfold ut_subst; fold ut_subst.
    rewrite ut_subst_comp.
    rewrite ut_church_subst.
    rewrite ut_succ_subst.
    ut_simpl.
    apply ut_beta_rt_rt.
    cut (S k' = nat_rec _ 0 (fun _ n => S n) (S k')).
    intros H; rewrite H at 2; clear H.
    apply ut_church_f_reduce.
    intro; apply ut_succ_prop.
    induction k'; simpl; f_equal; auto.
  Qed.
  
  Fact ut_apply_phi_solvable : ut_solvable ut_apply -> ut_solvable phi.
  Proof.
    intros H.
    apply ut_beta_eq_solvable with (1 := ut_apply_beta_eq) in H.
    revert H; apply ut_app_solvable.
  Qed.

End ut_apply.

Infix "#" := (@ut_apply _) (at level 61, left associativity).

Fact ut_apply_subst n phi v m (s : uts_t n m) : ut_subst (phi # v) s = ut_subst phi s # ut_subst v s.
Proof. unfold ut_apply; simpl; auto. Qed.

Section ut_apply_rec.

  Variables (n : nat).

  Fixpoint ut_apply_rev phi l : uterm n := 
    match l with 
      | nil  => phi
      | x::l => ut_apply_rev phi l # x
    end.
    
  Definition ut_apply_rec phi l := ut_apply_rev phi (rev l).

  Notation " f '#*' l " := (@ut_apply_rec f l) (at level 61, left associativity).
  Notation " f '*#' l " := (@ut_apply_rev f l) (at level 61, left associativity).

  Let ut_apply_rev_solvable phi l : Forall (@ut_value _) l -> ut_solvable (phi *# l) -> Forall (@ut_beta_eq_nat _) l.
  Proof.
    induction 1 as [ | x l Hx Hl IHl ]; simpl; intros H2.
    constructor.
    generalize (ut_apply_v_solvable H2); intros H3; apply Hx in H3.
    constructor; auto.
    clear Hx; destruct H3 as (u & Hx).
    generalize (ut_apply_beta_eq (ut_apply_rev phi l) Hx); intros H3.
    apply ut_beta_eq_solvable in H3.
    apply H3 in H2; clear H3.
    apply ut_app_solvable in H2.
    apply IHl; auto.
  Qed.
  
  (* This is Lemma 2.7 item 1 page 33 *)

  Fact ut_apply_rec_solvable phi l : Forall (@ut_value _) l -> ut_solvable (phi #* l) -> Forall (@ut_beta_eq_nat _) l.
  Proof.
    intros H1 H2.
    apply ut_apply_rev_solvable in H2.
    apply Forall_rev; auto.
    apply Forall_rev in H1; auto.
  Qed.

  Let ut_apply_rec_rev_prop_2 l : Forall (@ut_beta_eq_nat _) l
                            -> forall phi, phi *# l ~b phi *@ l.
  Proof.
    induction 1 as [ | x l [ v Hx ] Hl IHl ]; simpl; intros phi.
    constructor 2.
    apply ut_beta_eq_trans with (1 := ut_apply_beta_eq _ Hx).
    apply ut_beta_eq_app; auto.
    apply ut_beta_eq_sym; auto.
  Qed.
  
  (* This is Lemma 2.7 item 2 page 33 *)

  Fact ut_apply_rec_beta_eq l : Forall (@ut_beta_eq_nat _) l
                            -> forall phi, phi #* l ~b phi @* l.
  Proof.
    intros H phi.
    rewrite ut_app_rec_rev_eq.
    apply ut_apply_rec_rev_prop_2.
    apply Forall_rev in H; auto.
  Qed.

End ut_apply_rec.

Infix "#*" := (@ut_apply_rec _) (at level 61, left associativity).
Infix "*#" := (@ut_apply_rev _) (at level 61, left associativity).

Fact ut_apply_rev_subst n phi l m (s : uts_t n m) : ut_subst (phi *# l) s = ut_subst phi s *# map (fun v => ut_subst v s) l.
Proof.
  induction l as [ | x l IHl ]. simpl; auto.
  simpl ut_apply_rev.
  rewrite ut_apply_subst, IHl; auto.
Qed.

Fact ut_apply_rec_subst n phi l m (s : uts_t n m) : ut_subst (phi #* l) s = ut_subst phi s #* map (fun v => ut_subst v s) l.
Proof.
  unfold ut_apply_rec.
  rewrite ut_apply_rev_subst, map_rev; auto.
Qed.
