(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

(* This implements 

    "Strict Intersection Types for the Lambda Calculus" by Steffen van Bakel
 
   and Lambda Calculus, Types and Models by JL Krivine

*)

Require Import List Omega.

Require Import utils pos vec.

Require Import uterm uterm_eta uterm_beta uterm_beta_eq uterm_beta_eta.
Require Import uterm_normal.
Require Import itypes system_d uterm_sem.

Set Implicit Arguments.

Section semantic_results_for_system_D_bis.

  Variable X : Set. (* this is the ground type for intersection types *)
  
  Notation µ := (@in_stype_0 X).
  Infix "-->" := (@in_stype_1 X) (at level 60).
  Infix "<it" := (@le_it X) (at level 70).  

  Notation " e '|-' f ':w' t " := (@it_assign_wf X _ e f t) (at level 70).

  Section prop_3_19.

    Variable (n : nat) (p : pos n).

    Let N  (t : uterm n) := ut_beta_eta_sn t.
    Let N0 (t : uterm n) := exists x l, t = £ x *@ l /\ Forall N l.

    (* This one is the difficult one *)

    Let N_is_Nsat : Nsaturated N N.
    Proof.
      intros ? ? ?. 
      apply ut_beta_eta_sn_redex.
    Qed.

    Let N0_inc_N : N0 inc1 N.
    Proof.
      intros v (x & l & ? & H); subst.
      apply ut_beta_eta_sn_var_app_rev; auto.
    Qed.

    Let N0_N_Nadapted : Nadapted N N0.
    Proof.
      split; auto.
      split.
      intros t (x & l & ? & H) u Hu; subst.
      exists x, (u::l); split; auto.
      split.
      apply sem_arrow_mono; auto.
      intros u Hu.
      red in Hu.
      specialize (Hu (£ p)).
      inst Hu.
      exists p, nil; split; auto.
      revert Hu; apply ut_beta_eta_sn_left_inv.
    Qed.

    Let i : @sem X _ := fun _ => N.

    Let i_prop x : N0 inc1 i x /\ i x inc1 N /\ Nsaturated N (i x).
    Proof.
      unfold i; auto.
    Qed.

    Let K317 := @Krivine_3_17 _ _ _ _ _ N0_N_Nadapted i_prop.

    Theorem Krivine_3_20 e (f : uterm n) t : e |- f :w t::nil -> ut_beta_eta_sn f.
    Proof.
      intros H.

      generalize (@system_d_adequacy _ n (fun _ => N) n N e f (t::nil) (fun p => £ p) H).
      intros H1; inst H1.
      intros q x Hx.
      apply (ita_wf_hyps H q) in Hx.
      apply (@K317 _ Hx); exists q, nil; auto.
      inst H1; auto.
      inst H1.
      intros ? ?; apply K317; auto.
      specialize (H1 _ (or_introl eq_refl)).

      revert H1. 
      rewrite ut_subst_id.
      apply K317.
      apply (ita_wf_conc H); left; auto.
    Qed.

  End prop_3_19.

  Theorem typable_wf_implies_beta_eta_sn n (f : uterm n) : typable_w_free X f -> ut_beta_eta_sn f.
  Proof.
    intros (e & t & H).
    destruct t as [ | t l ].
    destruct (ita_wf_conc H) as [ [] ]; auto.
    rewrite ita_wf_int_inv in H.
    apply proj2 in H.
    specialize (H _ (or_introl eq_refl)).
    apply ut_beta_eta_lift_sn.
    rewrite <- ita_wf_lift with (l := t::nil) in H.
    revert H.
    apply Krivine_3_20 with (1 := pos0).
    apply (ita_wf_conc H).
  Qed.

End semantic_results_for_system_D_bis.

Check typable_wf_implies_beta_eta_sn.