(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import utils pos.

Require Import uterm uterm_beta uterm_beta_eq uterm_normal.

Set Implicit Arguments.

Fixpoint ut_beta_step n (f : uterm n) : { g | f -b-> g } + { ut_normal f }.
Proof.
  induction f as [ n f [ (g & IHf) | IHf ] 
                 | n f a Hf [ (f' & IHf) | IHf ] IHa
                 | n f a IHf IHa 
                 | n p ] using ut_custom_rect.
  left; exists (\\ g); apply in_ut_beta_abs, IHf.
  right; apply ut_abs_normal, IHf.
  left; exists (f' @ a); apply in_ut_beta_lft, IHf.
  destruct IHa as [ (a' & IHa) | IHa ].
  left; exists (f @ a'); apply in_ut_beta_rt, IHa.
  right; apply ut_app_normal; auto.
  left; exists (f[a]); apply in_ut_beta_redex.
  right; apply ut_var_normal.
Defined.

Definition ut_beta_step_b n (f : uterm n) : { g | f -b-> g } + { ut_normal_b f = true }.
Proof.
  destruct (ut_beta_step f) as [ H | H ].
  left; auto.
  right; apply ut_normal_b_prop, H.
Defined.

Definition ut_beta_not_normal_b n f : @ut_normal_b n f = false -> { g | f -b-> g }.
Proof.
  destruct (ut_beta_step_b f) as [ H | H ].
  intro; auto.
  rewrite H; discriminate 1.
Defined.

Definition ut_beta_one n f : @ut_normal_b n f = false -> uterm n.
Proof.
  intros H.
  destruct (ut_beta_not_normal_b f H); auto.
Defined.

Fact ut_beta_one_prop n f (H : @ut_normal_b n f = false) : f -b-> ut_beta_one f H.
Proof.
  apply (proj2_sig (ut_beta_not_normal_b f H)).
Qed.