(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith List Relations.

Require Import rel_utils utils cr pos.
Require Import uterm uterm_eta.

Set Implicit Arguments.

(* Definition ut_eta_or_eq n (u v : uterm n) := u -e-> v \/ u = v. *)

Notation " x 'e=' y " := (x -e-> y \/ x = y ) (at level 71).

Lemma ut_eta_square n f (a b : uterm n) : 
     f -e-> a -> f -e-> b -> exists g,  a e= g /\ b e= g.
Proof.
  intros H; revert H b.
  induction 1 as [ n f' a' | n f' h' a' H1 IH | n f' a' b' H1 IH | n f' h' H IH ];
    simpl; intros b Hb.

  (* redex *)

  apply ut_eta_abs_inv in Hb.
  destruct Hb as [ Hb | (w & ? & Hb) ]; subst.

  apply ut_app_inj, proj1, ut_lift_inj in Hb; subst f'.
  exists b; cbv; auto.

  apply ut_eta_app_inv in Hb.
  destruct Hb as [ (g & ? & Hb) | (g & ? & Hb) ]; subst.
  apply ut_eta_lift_inv in Hb.
  destruct Hb as (h & ? & Hb); subst.
  exists h; split; left; auto; constructor.

  apply ut_eta_var_inv in Hb.
  destruct Hb.

  (* left *)

  apply ut_eta_app_inv in Hb.
  destruct Hb as [ (g & ? & Hb) | (g & ? & Hb) ]; subst.
  destruct (IH _ Hb) as (u & H2 & H3).
  exists (u @ a').
  split.
  destruct H2 as [ | [] ]; [ left | right ]; auto; constructor 2; auto.
  destruct H3 as [ | [] ]; [ left | right ]; auto; constructor 2; auto.
  exists (h' @ g); split; left.
  constructor 3; auto.
  constructor 2; auto.

  (* right *)
  
  apply ut_eta_app_inv in Hb.
  destruct Hb as [ (g & ? & Hb) | (g & ? & Hb) ]; subst.
  exists (g @ a'); split; left.
  constructor 2; auto.
  constructor 3; auto.
  destruct (IH _ Hb) as (u & H2 & H3).
  exists (b' @ u).
  split.
  destruct H2 as [ | [] ]; [ left | right ]; auto; constructor 3; auto.
  destruct H3 as [ | [] ]; [ left | right ]; auto; constructor 3; auto.

  (* abs *)

  apply ut_eta_abs_inv in Hb.
  destruct Hb as [ Hb | (g & ? & Hb) ]; subst.

  apply ut_eta_app_inv in H.
  destruct H as [ (w & ? & H) | (w & ? & H) ]; subst.

  apply ut_eta_lift_inv in H.
  destruct H as (g & ? & H); subst.
  exists g; split; left; auto; constructor 1.

  apply ut_eta_var_inv in H; destruct H.

  destruct (IH _ Hb) as (h & H1 & H2).
  exists (\\ h); split.
  destruct H1; [ left; constructor 4; auto | right; f_equal; auto ].
  destruct H2; [ left; constructor 4; auto | right; f_equal; auto ].
Qed.

Theorem ut_eta_confluent n : confluent (@ut_eta n).
Proof.
  apply square_prop7.
  red; apply ut_eta_square.
Qed.